package edu.jhmi.telometer.calc.impl;

import edu.jhmi.telometer.bean.Cell;
import edu.jhmi.telometer.bean.Image;
import edu.jhmi.telometer.bean.ImageType;
import edu.jhmi.telometer.bean.LookupTables;
import edu.jhmi.telometer.bean.Mask;
import edu.jhmi.telometer.bean.Model;
import edu.jhmi.telometer.bean.NucleusEntry;
import edu.jhmi.telometer.bean.Pixelation;
import edu.jhmi.telometer.bean.Point;
import edu.jhmi.telometer.bean.Preset;
import edu.jhmi.telometer.bean.ScoreParams;
import edu.jhmi.telometer.bean.ScoreResults;
import edu.jhmi.telometer.bean.Scoring;
import edu.jhmi.telometer.bean.Stat;
import edu.jhmi.telometer.bean.Telomere;
import edu.jhmi.telometer.calc.api.GeneralUtil;
import edu.jhmi.telometer.calc.api.ImagePlusConvertUtil;
import edu.jhmi.telometer.calc.api.MaskConvertUtil;
import edu.jhmi.telometer.calc.api.MaskLogicUtil;
import edu.jhmi.telometer.calc.api.MaskRoiUtil;
import edu.jhmi.telometer.calc.api.ParticleUtil;
import edu.jhmi.telometer.calc.api.PixelModifier;
import edu.jhmi.telometer.calc.api.StatUtil;
import edu.jhmi.telometer.calc.api.StatUtilHelper;
import edu.jhmi.telometer.interfce.Named;
import edu.jhmi.telometer.interfce.Tagged;
import edu.jhmi.telometer.repo.CellRepo;
import edu.jhmi.telometer.repo.CellTypeRepo;
import edu.jhmi.telometer.repo.LesionTypeRepo;
import edu.jhmi.telometer.repo.PresetRepo;
import edu.jhmi.telometer.repo.ProjectRepo;
import edu.jhmi.telometer.repo.ScoringRepo;
import edu.jhmi.telometer.repo.TelomereRepo;
import edu.jhmi.telometer.repo.TissueTypeRepo;
import edu.jhmi.telometer.util.DoubleUtil;
import edu.jhmi.telometer.util.FileUtils;
import ij.ImagePlus;
import ij.gui.TextRoi;
import ij.process.ImageProcessor;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Component;
import org.springframework.util.Assert;

import java.awt.Rectangle;
import java.io.File;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.List;
import java.util.Set;
import java.util.stream.Collectors;

@Component
@Slf4j
public class GeneralUtilImpl implements GeneralUtil {

    @Autowired
    private ScoringRepo scoringRepo;
    @Autowired
    private CellRepo cellRepo;
    @Autowired
    private TelomereRepo telomereRepo;
    @Autowired
    private PresetRepo presetRepo;
    @Autowired
    private ProjectRepo projectRepo;
    @Autowired
    private TissueTypeRepo tissueTypeRepo;
    @Autowired
    private LesionTypeRepo lesionTypeRepo;
    @Autowired
    private CellTypeRepo cellTypeRepo;



    private ImagePlusConvertUtil imagePlusConvertUtil = new ImagePlusConvertUtilImpl();
    private MaskRoiUtil maskRoiUtil = new MaskRoiUtilImpl();
    private StatUtil statUtil = new StatUtilImpl();
    private StatUtilHelper statUtilHelper = new StatUtilHelperImpl();
    private MaskConvertUtil maskConvertUtil = new MaskConvertUtilImpl();
    private PixelModifier pixelModifier = new PixelModifierImpl();
    private MaskLogicUtil maskLogicUtil = new MaskLogicUtilImpl();
    private ParticleUtil particleUtil = new ParticleUtilImpl();


    /** Calculate the Cy Noise.  The cy image is supposed to stain only telomeres, but
     * it stains other things in the nucleus as well (it doesn't just attach to the telomeres).
     * The cynoise is defined as the average value of the cy image corresponding ONLY in the circled-cell in the DAPI.
     *
     * The cy noise is not the same thing as the overall background subtraction noise.
     * There is a part in the program where the user takes a square of black space in both the cy3 and dapi image.
     * This black space is represents no particles -- it could simply be some dust on the image.
     * This overall-background-subtraction represents noise for the ENTIRE IMAGE,
     * whereas the cy noise is noise only for the DAPI region of interest, and it only affects the pixels of the CY3 image.
     */
    public long calcCy3Noise1000(NucleusEntry nucleusEntry, Model model){
        Mask dapiMask = maskRoiUtil.roiToMask(nucleusEntry.getRoi());
        Image image = imagePlusConvertUtil.imagePlusToImage(model.getImagePlus(ImageType.CY3));
        Pixelation pixelation = maskConvertUtil.imageAndMaskToPixelation(image, dapiMask);
        Stat stat = statUtil.calcStat(pixelation);
        long cy3Noise1000 = stat.getMean1000(); //((double) stat.getSum()) / ((double) stat.getArea());
        return cy3Noise1000;
    }

    @Override
    public String findDefaultImageTag(){
        SimpleDateFormat dateFormat = new SimpleDateFormat("yyyyMMdd");
        Set<Named> nameds = scoringRepo.findAllNamesBy();
        Set<String> names = nameds.stream().map(t -> t.getName()).collect(Collectors.toSet());
        for (int i = 1; i < 9999; i++){
            String proposed = String.format("%s_%s", dateFormat.format(new Date()), i);
            if (!names.contains(proposed)){
                log.debug("proposed image tag: " + proposed);
                return proposed;
            }
        }
        log.error("Could not compute defaultImageTag");
        return "";
    }

    @Override
    public boolean isTagAlreadyUsed(String tag) {
        Set<Named> nameds = scoringRepo.findAllNamesBy();
        Set<String> tags = nameds.stream().map(t -> t.getName()).collect(Collectors.toSet());
        return tags.contains(tag);
    }




    @Override
    public ScoreResults calcScoreResults(ScoreParams scoreParams, Model model) {
        Scoring scoring = new Scoring();
        scoring.setProject(scoreParams.getProject());
        scoring.setName(scoreParams.getTag());
        scoring.setPreset(scoreParams.getPreset());
        scoring.setCy3Subtracted(scoreParams.isCy3Subtracted());
        scoring.setCyBackground(scoreParams.getCy3Background());
        scoring.setDapiBackground(scoreParams.getDapiBackground());
        scoring.setCreatedOn(System.currentTimeMillis());

        List<Cell> cells = new ArrayList();
        List<Telomere> telomeres = new ArrayList();
        for (NucleusEntry nucleusEntry : model.getNucleusHolder().getNucleusEntrys()){
            long cy3Noise1000 = calcCy3Noise1000(nucleusEntry, model);
            Cell cell = new Cell();
            {
                Mask dapiMask = maskRoiUtil.roiToMask(nucleusEntry.getRoi());
                ImagePlus imagePlus = model.getImagePlus(ImageType.DAPI);
                Image image = imagePlusConvertUtil.imagePlusToImage(imagePlus);
                Pixelation pixelation = maskConvertUtil.imageAndMaskToPixelation(image, dapiMask);
                Stat stat = statUtil.calcStat(pixelation);
                cell.setScoring(scoring);
                cell.setName(nucleusEntry.getName());
                cell.setLesionType(nucleusEntry.getLesionType());
                cell.setTissueType(nucleusEntry.getTissueType());
                cell.setCellType(nucleusEntry.getCellType());
                cell.setCellCount(nucleusEntry.getCellCount());
                cell.setNotes(nucleusEntry.getNotes());
                cell.setMin(stat.getMin());
                cell.setMax(stat.getMax());
                cell.setMean1000(stat.getMean1000());
                cell.setSum(stat.getSum());
                cell.setArea(stat.getArea());
                cell.setStddev1000(stat.getStddev1000());
                cell.setCy3Noise1000(cy3Noise1000);
                Rectangle rectangle = nucleusEntry.getRoi().getBounds();
                cell.setWidth((int) Math.round(rectangle.getWidth()));
                cell.setHeight((int) Math.round(rectangle.getHeight()));
                cell.setX((int) Math.round(rectangle.getX()));
                cell.setY((int) Math.round(rectangle.getY()));
                cells.add(cell);
            }
            {
                ImagePlus imagePlus = model.getImagePlus(ImageType.COMBINED);
                Image image = imagePlusConvertUtil.imagePlusToImage(imagePlus);
                Mask combinedMask = maskConvertUtil.nonZeroImageToMask(image);
                Mask dapiMask = maskRoiUtil.roiToMask(nucleusEntry.getRoi());
                Mask mask = maskLogicUtil.andMasks(combinedMask, dapiMask);
                Set<Mask> particleMasks = particleUtil.breakImageMaskIntoParticles(mask);
                int telomereNumber = 0;
                int cy3Noise = (int) Math.round(DoubleUtil.long1000ToDouble(cy3Noise1000));
                Preset preset = scoreParams.getPreset();
                for (Mask particleMask : particleMasks){
                    Pixelation pixelation = maskConvertUtil.imageAndMaskToPixelation(image, particleMask);
                    if (scoreParams.isCy3Subtracted()) {
                        pixelation = pixelModifier.subtractIntensity(pixelation, cy3Noise);
                    }
                    Stat stat = statUtil.calcStat(pixelation);
                    int area = stat.getArea();
                    if (area < preset.getMinSpotSize() || area > preset.getMaxSpotSize()){
                        continue;
                    }

                    Point averagePoint = maskConvertUtil.averagePoint(particleMask);

                    Telomere telomere = new Telomere();
                    telomere.setCell(cell);
                    ++telomereNumber;
                    telomere.setTelomereNumber(telomereNumber);
                    telomere.setArea(area);
                    telomere.setMin(stat.getMin());
                    telomere.setMax(stat.getMax());
                    telomere.setStddev1000(stat.getStddev1000());
                    telomere.setMean1000(stat.getMean1000());
                    telomere.setSum(stat.getSum());
                    telomere.setX(averagePoint.getX());
                    telomere.setY(averagePoint.getY());
                    telomeres.add(telomere);
                }
            }
        }
        ScoreResults scoreResults = ScoreResults.builder()
                .scoring(scoring)
                .cells(cells)
                .telomeres(telomeres)
                .build();

        logScoreResults(scoreResults);
        return scoreResults;
    }

    @Override
    public void dumpScoreResultsToDatabase(ScoreResults scoreResults) {
        scoringRepo.save(scoreResults.getScoring());
        for (Cell cell : scoreResults.getCells()){
            cellRepo.save(cell);
        }
        for (Telomere telomere : scoreResults.getTelomeres()){
            telomereRepo.save(telomere);
        }
        log.info("Scoring Results saved to database. Saved Details follow.");
        logScoreResults(scoreResults);
        verifyScoreResults(scoreResults);
    }

    private static void verifyScoreResults(ScoreResults scoreResults) {
        Assert.notNull(scoreResults.getScoring().getId(), "scoring was never saved");
        for (Cell cell : scoreResults.getCells()){
            Assert.notNull(cell.getId(), "cell was never saved");
        }
        for (Telomere telomere : scoreResults.getTelomeres()){
            Assert.notNull(telomere.getId(), "telomere was never saved");
        }
    }

    private void logScoreResults(ScoreResults scoreResults){
        log.info("Calculations complete. Unsaved Details follow.");
        log.debug("scoring: {}", scoreResults.getScoring());
        for (Cell cell : scoreResults.getCells()){
            log.debug("cell: {}", cell);
        }
        for (Telomere telomere : scoreResults.getTelomeres()){
            log.debug("telomere: {}", telomere);
        }
    }

    @Override
    public LookupTables lookupTables() {
        LookupTables lookupTables = LookupTables.builder()
                .presets(presetRepo.findAll())
                .projects(projectRepo.findAll())
                .tissueTypes(tissueTypeRepo.findAll())
                .cellTypes(cellTypeRepo.findAll())
                .lesionTypes(lesionTypeRepo.findAll())
                .build();
        return lookupTables;
    }

}
