package edu.jhmi.telometer.calc.impl;

import edu.jhmi.telometer.bean.Dimension;
import edu.jhmi.telometer.bean.Image;
import edu.jhmi.telometer.bean.Pixelation;
import edu.jhmi.telometer.calc.api.ImagePlusConvertUtil;
import ij.ImagePlus;
import ij.io.TiffEncoder;
import ij.process.ByteProcessor;
import ij.process.ColorProcessor;
import ij.process.FloatProcessor;
import ij.process.ImageProcessor;
import ij.process.ShortProcessor;
import lombok.extern.slf4j.Slf4j;
import org.springframework.util.Assert;

import java.io.ByteArrayOutputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

@Slf4j
public class ImagePlusConvertUtilImpl implements ImagePlusConvertUtil {

    @Override
    public Image imagePlusToImage(ImagePlus imagePlus) {
        return Image.builder()
                .dimension(imagePlusToDimension(imagePlus))
                .pixelation(imagePlusToPixelation(imagePlus))
                .build();
    }

    private Pixelation imagePlusToPixelation(ImagePlus imagePlus) {
        ImageProcessor imageProcessor = imagePlus.getProcessor();
        Assert.isTrue(imageProcessor instanceof ShortProcessor, "must be a shortProcessor to get pixels");
        ShortProcessor shortProcessor = (ShortProcessor) imageProcessor;
        short[] pixels = (short[]) shortProcessor.getPixels();
        List<Integer> pixelInts = new ArrayList();
        for (short s : pixels){
            int intValue = shortToIntValue(s);
            pixelInts.add(intValue);
        }
        Pixelation pixelation = Pixelation.builder().pixels(pixelInts).build();
        return pixelation;
    }

    private static int shortToIntValue(short s) {
        return s & 0xFFFF;
    }

    private static short intToShortValue(int i){
        Assert.isTrue((i >= 0) && (i <= 65536), "i is outOfBounds for a short: " + i);
        int newInt = i > Short.MAX_VALUE ? -65536 + i : i;
        short returned = (short) newInt;
        return returned;
    }

    @Override
    public Dimension imagePlusToDimension(ImagePlus imagePlus) {
        return Dimension.builder().width(imagePlus.getWidth()).height(imagePlus.getHeight()).build();
    }

    @Override
    public short[] pixelationToShortArray(Pixelation pixelation) {
        List<Integer> pixelInts = pixelation.getPixels();
        short[] pixels = new short[pixelInts.size()];
        for (int i = 0; i < pixelInts.size(); i++){
            int intValue = pixelInts.get(i);
            short shortValue = intToShortValue(intValue);
            pixels[i] = shortValue;
        }
        return pixels;
    }

    @Override
    public byte[] convertImagePlusToTiffBytes(ImagePlus imagePlus) {
        TiffEncoder tiffEncoder = new TiffEncoder(imagePlus.getFileInfo());
        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        DataOutputStream dos = new DataOutputStream(baos);
        try {
            tiffEncoder.write(dos);
        } catch (IOException e) {
            String message = "Could not convert Image Plus to tiff bytes: " + e;
            throw new RuntimeException(message, e);
        }
        byte[] bytes = baos.toByteArray();
        return bytes;
    }

    @Override
    public String getReasonImagePlusNotAcceptable(ImagePlus imagePlus) {
        String reason = getWrongImageTypeDescription(imagePlus);
        if (reason == null){
            return reason;
        }
        final String URL = "https://demarzolab.pathology.jhmi.edu/telometer/faq.html#imageType";
        String returned = String.format("Image needs to be a 16-bit grayscale, but it is a %s image. See:\n%s", reason, URL);
        return returned;
    }


    private String getWrongImageTypeDescription(ImagePlus imagePlus) {
        if (imagePlus == null){
            return "non";
        }
        ImageProcessor processor = imagePlus.getProcessor();
        if (processor instanceof ColorProcessor){
            return "color";
        }
        if (processor instanceof ByteProcessor){
            return "8-bit";
        }
        if (processor instanceof FloatProcessor){
            return "32-bit";
        }
        if (! (processor instanceof ShortProcessor)){
            return "non 16-bit";
        }
        return null;
    }



}
