package edu.jhmi.telometer.util;

import edu.jhmi.telometer.bean.Cell;
import edu.jhmi.telometer.bean.CellType;
import edu.jhmi.telometer.bean.LesionType;
import edu.jhmi.telometer.bean.Preset;
import edu.jhmi.telometer.bean.Project;
import edu.jhmi.telometer.bean.Scoring;
import edu.jhmi.telometer.bean.Telomere;
import edu.jhmi.telometer.bean.TissueType;
import edu.jhmi.telometer.repo.CellRepo;
import edu.jhmi.telometer.repo.CellTypeRepo;
import edu.jhmi.telometer.repo.LesionTypeRepo;
import edu.jhmi.telometer.repo.PresetRepo;
import edu.jhmi.telometer.repo.ProjectRepo;
import edu.jhmi.telometer.repo.ScoringRepo;
import edu.jhmi.telometer.repo.TelomereRepo;
import edu.jhmi.telometer.repo.TissueTypeRepo;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;
import org.springframework.util.Assert;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.stream.Collectors;

@Component
@Slf4j
public class DbPopulator {

    public static final String UNSPECIFIED = "unspecified";


    @Autowired
    private ProjectRepo projectRepo;

    @Autowired
    private PresetRepo presetRepo;

    @Autowired
    private CellTypeRepo cellTypeRepo;

    @Autowired
    private LesionTypeRepo lesionTypeRepo;

    @Autowired
    private TissueTypeRepo tissueTypeRepo;

    @Autowired
    private ScoringRepo scoringRepo;

    @Autowired
    private CellRepo cellRepo;

    @Autowired
    private TelomereRepo telomereRepo;

    private static final String[] PRESET_NAMES = {
            "100X (telomere)",
            "60X (telomere)",
            "40X (telomere)",
            "60X (centromere)",
            "40X (centromere)",
    };
    private static final int[] minSpotSizes = {10, 4, 0, 5, 1};
    private static final int[] maxSpotSizes = {350, 200, 500, 5000, 700};
    private static final int[] despeckleRatio1000s = {500, 400, 200, 400, 400};
    private static final int[] rollingBallSizes = {5, 2, 1, 9, 2};


    private static final String[] tissueTypes = {
            UNSPECIFIED,
            "aorta",
            "bladder",
            "blood",
            "brain",
            "breast",
            "colon",
            "heart",
            "kidney",
            "lung",
            "lymph node",
            "ovary",
            "prostate",
            "pancreas",
            "placenta",
            "testis",
            "uterus",
    };

    private static final String[] lesionTypes = {
            UNSPECIFIED,
            "atrophy",
            "benign",
            "cancer",
            "hgpin",
    };

    private static final String[] cellTypes = {
            UNSPECIFIED,
            "acinar cell",
            "ductal cell",
            "epythelial",
            "fibroblast",
            "luminal",
            "stromal"
    };

    private static final String[] projects = {
            "my project",
            "test",
    };


    private static List<Preset> presets(){
        List<Preset> presets = new ArrayList();
        for (int i = 0; i < PRESET_NAMES.length; i++){
            presets.add(Preset.builder()
                    .name(PRESET_NAMES[i])
                    .minSpotSize(minSpotSizes[i])
                    .maxSpotSize(maxSpotSizes[i])
                    .maxSpotSize(maxSpotSizes[i])
                    .despeckleRatio1000(despeckleRatio1000s[i])
                    .rollingBallSize(rollingBallSizes[i])
                    .build());
        }
        return presets;
    }

    private List<CellType> cellTypes(){
        return Arrays.stream(cellTypes)
                .map(t -> CellType.builder().name(t).build())
                .collect(Collectors.toList());
    }

    private List<LesionType> lesionTypes(){
        return Arrays.stream(lesionTypes)
                .map(t -> LesionType.builder().name(t).build())
                .collect(Collectors.toList());
    }


    private List<TissueType> tissueTypes(){
        return Arrays.stream(tissueTypes)
                .map(t -> TissueType.builder().name(t).build())
                .collect(Collectors.toList());
    }

    private List<Project> projects(){
        return Arrays.stream(projects)
                .map(t -> Project.builder().name(t).build())
                .collect(Collectors.toList());
    }

    public void removeEverything(){
        log.info("removing data from all lookup tables");
        telomereRepo.deleteAll();
        cellRepo.deleteAll();
        scoringRepo.deleteAll();
        projectRepo.deleteAll();
        presetRepo.deleteAll();
        cellTypeRepo.deleteAll();
        lesionTypeRepo.deleteAll();
        tissueTypeRepo.deleteAll();
    }

    public boolean areLookupTablesPopulated(){
        boolean areLookupTablesPopulated = !presetRepo.findAll().isEmpty();
        log.debug("lookup tables {} already populated", areLookupTablesPopulated ? "are" : "aren't");
        return areLookupTablesPopulated;
    }

    public void populateDbLookupTables(){
        Assert.isTrue(presetRepo.findAll().isEmpty(), "database already populated");
        log.info("populating lookup tables with hard-coded data");
        presetRepo.saveAll(presets());
        cellTypeRepo.saveAll(cellTypes());
        lesionTypeRepo.saveAll(lesionTypes());
        tissueTypeRepo.saveAll(tissueTypes());
        projectRepo.saveAll(projects());
    }

    public void populateDbWithTestData(){
        log.info("populating database with test data");
        Preset preset = presetRepo.findAll().get(0);
        TissueType tissueType = tissueTypeRepo.findAll().get(0);
        LesionType lesionType = lesionTypeRepo.findAll().get(0);
        CellType cellType = cellTypeRepo.findAll().get(0);
        List<Project> projects = projectRepo.findAll();
        for (int i = 0; i < 5; i++){
            Scoring scoring = Scoring.builder()
                    .preset(preset)
                    .project(projects.get(i % projects.size()))
                    .name("round#" + i)
                    .build();
            scoringRepo.save(scoring);
            for (int j = 0; j < 3; j++){
                Cell cell = Cell.builder()
                        .scoring(scoring)
                        .cellType(cellType)
                        .lesionType(lesionType)
                        .tissueType(tissueType)
                        .name("cell#" + j)
                        .build();
                cellRepo.save(cell);
                for (int k = 0; k < 4; k++){
                    Telomere telomere = Telomere.builder()
                            .telomereNumber(k + 1)
                            .cell(cell)
                            .build();
                    telomereRepo.save(telomere);
                }
            }
        }
    }
}
