package edu.jhmi.telometer.util.impl;

import edu.jhmi.telometer.bean.ImageType;
import edu.jhmi.telometer.bean.Mask;
import edu.jhmi.telometer.bean.Model;
import edu.jhmi.telometer.bean.NucleusEntry;
import edu.jhmi.telometer.bean.Point;
import edu.jhmi.telometer.bean.Scoring;
import edu.jhmi.telometer.calc.api.ImagePlusConvertUtil;
import edu.jhmi.telometer.calc.api.MaskConvertUtil;
import edu.jhmi.telometer.calc.api.MaskRoiUtil;
import edu.jhmi.telometer.calc.api.PixelModifier;
import edu.jhmi.telometer.calc.impl.ImagePlusConvertUtilImpl;
import edu.jhmi.telometer.calc.impl.MaskConvertUtilImpl;
import edu.jhmi.telometer.calc.impl.MaskRoiUtilImpl;
import edu.jhmi.telometer.calc.impl.PixelModifierImpl;
import edu.jhmi.telometer.util.FileUtils;
import edu.jhmi.telometer.util.api.ImageDirUtil;
import ij.ImagePlus;
import ij.gui.TextRoi;
import ij.process.ImageProcessor;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Component;
import org.springframework.util.Assert;

import java.io.File;
import java.text.NumberFormat;
import java.util.Arrays;

@Component
@Slf4j
public class ImageDirUtilImpl implements ImageDirUtil {

    @Value("${data.dir}")
    private String dataDirName;

    private ImagePlusConvertUtil imagePlusConvertUtil = new ImagePlusConvertUtilImpl();
    private PixelModifier pixelModifier = new PixelModifierImpl();
    private MaskConvertUtil maskConvertUtil = new MaskConvertUtilImpl();
    private MaskRoiUtil maskRoiUtil = new MaskRoiUtilImpl();

    private static NumberFormat numberFormat = NumberFormat.getNumberInstance();

    static {
        numberFormat.setMinimumFractionDigits(0);
        numberFormat.setMaximumFractionDigits(0);
        numberFormat.setMinimumIntegerDigits(5);
        numberFormat.setMaximumIntegerDigits(5);
        numberFormat.setGroupingUsed(false);
    }

    @Override
    public void dumpImages(Scoring scoring, Model model) {
        log.info("dumping images to directory for scoring#{}", scoring.getId());
        for (ImageType imageType : ImageType.values()) {
            ImagePlus imagePlus = model.getImagePlus(imageType);
            writeImagePlusToDir(imagePlus, imageType, scoring);
        }
    }

    private void writeImagePlusToDir(ImagePlus imagePlus, ImageType imageType, Scoring scoring) {
        byte[] bytes = imagePlusConvertUtil.convertImagePlusToTiffBytes(imagePlus);
        File file = getFileForImageType(scoring, imageType);
        log.debug("writing image file for scoring#{} for imageType{} in file {}",
                scoring.getId(), imageType, file.getAbsolutePath());
        FileUtils.writeBytesToFile(bytes, file);
    }

    public File getFileForImageType(Scoring scoring, ImageType imageType){
        File imageDir = getScoringImageDir(scoring);
        String fileName = String.format("%s_%s.tif", imageType.toString().toLowerCase(),
                scoring.getId());
        File file = new File(imageDir, fileName);
        return file;
    }

    @Override
    public ImagePlus getAnnotatedImagePlus(Model model) {
        ImagePlus imagePlus = copyImagePlus(model.getImagePlus(ImageType.DAPI));
        for (NucleusEntry nucleusEntry : model.getNucleusHolder().getNucleusEntrys()) {
            pixelModifier.applyRoiToImagePlus(nucleusEntry.getRoi(), imagePlus);
            TextRoi textRoi = nucleusEntryToTextRoi(nucleusEntry);
            pixelModifier.applyRoiToImagePlus(textRoi, imagePlus);
        }
        return imagePlus;
    }

    private TextRoi nucleusEntryToTextRoi(NucleusEntry nucleusEntry) {
        Mask mask = maskRoiUtil.roiToMask(nucleusEntry.getRoi());
        Point point = maskConvertUtil.averagePoint(mask);
        TextRoi textRoi = new TextRoi(point.getX(), point.getY(), nucleusEntry.getName());
        return textRoi;

    }

    private static ImagePlus copyImagePlus(ImagePlus srcImagePlus) {
        ImageProcessor srcProcessor = srcImagePlus.getProcessor();
        short[] srcPixels = (short[]) srcProcessor.getPixels();
        short[] dstPixels = Arrays.copyOf(srcPixels, srcPixels.length);
        ImageProcessor dstProcessor = srcProcessor.createProcessor(srcImagePlus.getWidth(), srcImagePlus.getHeight());
        dstProcessor.setPixels(dstPixels);
        ImagePlus dstImagePlus = new ImagePlus("copy", dstProcessor);
        return dstImagePlus;
    }


    private File getScoringImageDir(Scoring scoring) {
        File middleDir = getMiddleDir(scoring);
        String name = getScoringDirName(scoring);
        File dir = new File(middleDir, name);
        if (!dir.exists()){
            log.info("Creating scoringImageDir: {}", dir.getAbsolutePath());
            dir.mkdirs();
        }
        return dir;
    }

    private String getScoringDirName(Scoring scoring){
        Long id = scoring.getId();
        Assert.notNull(id, "scoring id is null");
        return String.valueOf(id);
    }

    private File getMiddleDir(Scoring scoring){
        Long id = scoring.getId();
        Assert.notNull(id, "scoring id is null");
        File baseImageDir = getBaseImageDir();
        String name = getMiddleScoringImageDirName(id);
        File dir = new File(baseImageDir, name);
        dir.mkdirs();
        return dir;
    }

    /**
     * 0-99 :   00000
     * 100-199: 00100
     * 200-299: 00200
     */
    private String getMiddleScoringImageDirName(long scoringId) {
        scoringId /= 100;
        String name = numberFormat.format(scoringId);
        return name;
    }

    private File getBaseImageDir() {
        File dataDir = getDataDir();
        File dir = new File(dataDir, "images");
        if (!dir.exists()) {
            log.info("Creating baseImageDir: {}", dir.getAbsolutePath());
            dir.mkdirs();
        }
        return dir;
    }

    @Override
    public File getDataDir(){
        try {
            File dir = new File(dataDirName);
            if (!dir.exists()) {
                log.info("Creating data directory: " + dir.getAbsolutePath());
                dir.mkdirs();
            }
            return dir;
        } catch (RuntimeException e){
            log.error("Could not create dataDir", e);
            throw new RuntimeException(e);
        }
    }

    @Override
    public void removeImages(Scoring scoring) {
        log.info("removing images for scoring#{}", scoring.getId());
        for (ImageType imageType : ImageType.values()) {
            File file = getFileForImageType(scoring, imageType);
            deleteFile(file);
        }
        deleteScoringDir(scoring);
    }

    private void deleteScoringDir(Scoring scoring) {
        File dir = getScoringImageDir(scoring);
        if (!dir.isDirectory()){
            log.error("Can't delete scoring directory.  not a directory: {}", dir.getAbsolutePath());
            return;
        }
        if (dir.listFiles().length > 0){
            log.error("Can't delete scoring directory.  not empty: {}", dir.getAbsolutePath());
            return;
        }
        if (!isInPath(getBaseImageDir(), dir)){
            log.error("Can't delete scoring dir.  Dir isn't inside imageDir '{}': {}", getBaseImageDir().getAbsolutePath(), dir.getAbsolutePath());
            return;
        }
        log.info("deleting directory for scoringId{}: {}", scoring.getId(), dir.getAbsolutePath());
        dir.delete();
    }

    private void deleteFile(File file) {
        if (!file.isFile()){
            log.error("Can't delete image.  Not a file: {}", file.getAbsolutePath());
            return;
        }
        if (!isValidFileNameExtension(file.getName())){
            log.error("Can't delete image.  File doesn't have a valid image fileName extension: {}", file.getAbsolutePath());
            return;
        }
        if (!isValidFileNameName(file.getName())){
            log.error("Can't delete image.  File doesn't have a valid imageType in name: {}", file.getAbsolutePath());
            return;
        }
        if (!isInPath(getBaseImageDir(), file)){
            log.error("Can't delete image.  File isn't inside imageDir '{}': {}", getBaseImageDir().getAbsolutePath(), file.getAbsolutePath());
            return;
        }
        log.info("deleting file: {}", file.getAbsolutePath());
        boolean wasDeleted = file.delete();
        if (!wasDeleted){
            log.error("Could not delete image file: {}", file.getAbsolutePath());
            return;
        }
    }



    static boolean isValidFileNameExtension(String fileName){
        String[] EXTENSIONS = {
                "bmp", "jpg", ".jpeg", "tif", "tiff",
                "png", "gif"
        };
        fileName = fileName.toLowerCase();
        for (String extension: EXTENSIONS){
            if (fileName.endsWith("." + extension)){
                return true;
            }
        }
        return false;
    }

    static boolean isValidFileNameName(String fileName){
        fileName = fileName.toLowerCase();
        for (ImageType imageType : ImageType.values()){
            if (fileName.contains(imageType.toString().toLowerCase())){
                return true;
            }
        }
        return false;
    }

    static boolean isInPath(File main, File under){
        if (under == null){
            return false;
        }
        if (main.equals(under)){
            return true;
        }
        return isInPath(main, under.getParentFile());
    }
}
