package edu.jhmi.telometer.view.admin;

import edu.jhmi.telometer.bean.Preset;
import edu.jhmi.telometer.interfce.Dialogable;
import edu.jhmi.telometer.interfce.OkListener;
import edu.jhmi.telometer.thirdparty.SpringUtilities;
import edu.jhmi.telometer.util.DialogUtil;
import edu.jhmi.telometer.util.DoubleUtil;

import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JDialog;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JSpinner;
import javax.swing.JTextField;
import javax.swing.SpinnerNumberModel;
import javax.swing.SpringLayout;
import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Frame;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

public class PresetDialog implements Dialogable {

    private static final int SPOT_SIZE_MINIMUM = 0;
    private static final int SPOT_SIZE_MAXIMUM = 99999;
    private static final int ROLLING_BALL_SIZE_MINIMUM = 1;
    private static final int ROLLING_BALL_SIZE_MAXIMUM = 999;
    private static final double MIN_DESPECKLE_RATIO = 0.001;
    private static final double MAX_DESPECKLE_RATIO = 1.000;

    private OkListener<Preset> okListener;


    private final JTextField idTextField = new JTextField();
    private final JTextField nameTextField = new JTextField("myPreset");
    private final JSpinner minSpotSizeSpinner = new JSpinner(new SpinnerNumberModel(10, SPOT_SIZE_MINIMUM, SPOT_SIZE_MAXIMUM, 1));
    private final JSpinner maxSpotSizeSpinner = new JSpinner(new SpinnerNumberModel(350, SPOT_SIZE_MINIMUM, SPOT_SIZE_MAXIMUM, 1));
    private final JSpinner rollingBallSizeSpinner = new JSpinner(new SpinnerNumberModel(5, ROLLING_BALL_SIZE_MINIMUM, ROLLING_BALL_SIZE_MAXIMUM, 1));
    private final JSpinner despeckleRatioSpinner = new JSpinner(new SpinnerNumberModel(0.500, MIN_DESPECKLE_RATIO, MAX_DESPECKLE_RATIO, 0.010));
    private final JPanel mainPanel = new JPanel(new BorderLayout());
    private final JButton okButton = new JButton("Ok");
    private final JButton cancelButton = new JButton("Cancel");
    private final JDialog dialog;


    public PresetDialog(Component component, Preset preset, boolean isReferenced){
        setToolTips();
        idTextField.setEditable(false);
        if (isReferenced){
            minSpotSizeSpinner.setEnabled(false);
            maxSpotSizeSpinner.setEnabled(false);
            rollingBallSizeSpinner.setEnabled(false);
            despeckleRatioSpinner.setEnabled(false);
        }
        idTextField.setBackground(Color.LIGHT_GRAY);
        mainPanel.add(getNorthPanel(preset), BorderLayout.NORTH);
        mainPanel.add(getSouthPanel(isReferenced), BorderLayout.SOUTH);
        mainPanel.add(getFormPanel(), BorderLayout.CENTER);

        idTextField.setText(preset.getId() != null ? preset.getId() + "" : "");
        nameTextField.setText(preset.getName());
        minSpotSizeSpinner.setValue(preset.getMinSpotSize());
        maxSpotSizeSpinner.setValue(preset.getMaxSpotSize());
        rollingBallSizeSpinner.setValue(preset.getRollingBallSize());
        despeckleRatioSpinner.setValue(DoubleUtil.long1000ToDouble(preset.getDespeckleRatio1000()));


        okButton.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent actionEvent) {
                okayButtonClicked();
            }
        });

        Frame frame = JOptionPane.getFrameForComponent(component);
        dialog = new JDialog(frame, "Preset", true);
        cancelButton.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent actionEvent) {
                dialog.dispose();
            }
        });
        dialog.getContentPane().add(mainPanel);
        dialog.pack();
    }

    private void setToolTips() {
        idTextField.setToolTipText("the internal id assigned in the database (not editable)");
        nameTextField.setToolTipText("the name displayed on the scoring options page");
        rollingBallSizeSpinner.setToolTipText("Set the size of the 'rolling ball background subtractor' See: https://imagej.net/Rolling_Ball_Background_Subtraction");

        String despeckle = "<HTML>" +
                "A despeckle filter is run after immediately after the thresholding operation and will  <BR/>" +
                "clean up most of the leftover noise on the binary image.  <BR/>" +
                "It replaces each pixel with the median value in its 3 × 3 neighborhood.  <BR/>" +
                "This ratio controls the percentage of 3 x 3 pixels on the binary image that must be 'white' <BR/>" +
                "in order to call the median pixel white.  Otherwise the median pixel is set as black. </HTML>";
        despeckleRatioSpinner.setToolTipText(despeckle);
        maxSpotSizeSpinner.setToolTipText("any particles with pixel-area higher than this number will be excluded");
        minSpotSizeSpinner.setToolTipText("any particles with pixel-area lower than this number will be excluded");
    }


    private void okayButtonClicked() {
        String id = idTextField.getText();
        String name = nameTextField.getText();
        String despeckleRatio = despeckleRatioSpinner.getValue().toString();
        String minSpotSize = minSpotSizeSpinner.getValue().toString();
        String maxSpotSize = maxSpotSizeSpinner.getValue().toString();
        String rollingBallSize = rollingBallSizeSpinner.getValue().toString();

        if (!isInteger(minSpotSize)){
            showMessage("minSpotSize must be an int");
            return;
        }
        if (!isInteger(maxSpotSize)){
            showMessage("maxSpotSize must be an int");
            return;
        }
        if (!isInteger(rollingBallSize)){
            showMessage("rollingBallSize must be an int");
            return;
        }
        if (!isFloat(despeckleRatio)){
            showMessage("despeckleRatio must be a float");
            return;
        }
        int minSpot = Integer.parseInt(minSpotSize);
        int maxSpot = Integer.parseInt(maxSpotSize);
        if (minSpot > maxSpot){
            showMessage("minSpot can't be larger than maxSpot");
            return;
        }


        Preset preset = Preset.builder()
                .id(isInteger(id) ? Long.parseLong(id) : null)
                .name(name)
                .minSpotSize(minSpot)
                .maxSpotSize(maxSpot)
                .rollingBallSize(Integer.parseInt(rollingBallSize))
                .despeckleRatio1000(DoubleUtil.doubleToLong1000(Double.parseDouble(despeckleRatio)))
                .build();
        String presetErrorMessage = getPresetErrorMessage(preset);
        if (presetErrorMessage != null){
            showMessage(presetErrorMessage);
            return;
        }
        okListener.okayButtonClicked(preset);
    }

    private String getPresetErrorMessage(Preset preset) {
        if (preset.getName().trim().equals("")){
            return "name must have a value";
        }
        if (preset.getMinSpotSize() < SPOT_SIZE_MINIMUM || preset.getMinSpotSize() > SPOT_SIZE_MAXIMUM){
            return String.format("minSpotSize must be between %s and %s and was '%s'", SPOT_SIZE_MINIMUM, SPOT_SIZE_MAXIMUM, preset.getMinSpotSize());
        }
        if (preset.getMaxSpotSize() < SPOT_SIZE_MINIMUM || preset.getMaxSpotSize() > SPOT_SIZE_MAXIMUM){
            return String.format("maxSpotSize must be between %s and %s and was '%s'", SPOT_SIZE_MINIMUM, SPOT_SIZE_MAXIMUM, preset.getMaxSpotSize());
        }
        if (preset.getRollingBallSize() < ROLLING_BALL_SIZE_MINIMUM || preset.getRollingBallSize() > ROLLING_BALL_SIZE_MAXIMUM){
            return String.format("rollingBallSize must be between %s and %s and was '%s'", ROLLING_BALL_SIZE_MINIMUM, ROLLING_BALL_SIZE_MINIMUM, preset.getMaxSpotSize());
        }
        double despeckleRatio = DoubleUtil.long1000ToDouble(preset.getDespeckleRatio1000());
        if (isDespeckleRatioBad(despeckleRatio)){
            return getDespeckleErrorMessage(despeckleRatio);
        }
        return null;
    }

    public static boolean isDespeckleRatioBad(double despeckleRatio){
        return (despeckleRatio < MIN_DESPECKLE_RATIO || despeckleRatio > MAX_DESPECKLE_RATIO);
    }

    public static String getDespeckleErrorMessage(double despeckleRatio){
        return String.format("despeckleRatio must be between %s and %s and was '%s'", MIN_DESPECKLE_RATIO, MAX_DESPECKLE_RATIO, despeckleRatio);
    }


    private JPanel getNorthPanel(Preset preset) {
        JPanel panel = new JPanel();
        String text = String.format("%s Preset", preset.getId() == null ? "Create" : "Edit");
        panel.add(new JLabel(text));
        return panel;
    }

    private JPanel getSouthPanel(boolean isReferenced){
        JPanel northPanel = new JPanel();
        if (isReferenced){
            northPanel.add(new JLabel("<HTML><SMALL><I>* If a preset has been using by a scoring, only its name can be changed</I></SMALL></HTML>"));
        }
        JPanel southPanel = new JPanel();
        southPanel.add(okButton);
        southPanel.add(cancelButton);
        JPanel panel = new JPanel(new BorderLayout());
        panel.add(northPanel, BorderLayout.NORTH);
        panel.add(southPanel, BorderLayout.SOUTH);
        return panel;
    }


    private JPanel getFormPanel() {
        JPanel formPanel = new JPanel(new SpringLayout());

        JLabel[] labels = {
                new JLabel("ID"),
                new JLabel("Name"),
                new JLabel("Minimum Spot Size"),
                new JLabel("Maximum Spot Size"),
                new JLabel("Rolling Ball Size"),
                new JLabel("Despeckle Ratio"),
        };

        JComponent[] components = {
                idTextField,
                nameTextField,
                minSpotSizeSpinner,
                maxSpotSizeSpinner,
                rollingBallSizeSpinner,
                despeckleRatioSpinner
        };
        for (int i = 0; i < labels.length; i++) {
            formPanel.add(labels[i]);
            formPanel.add(components[i]);
            labels[i].setToolTipText(components[i].getToolTipText());
        }
        SpringUtilities.makeCompactGrid(formPanel,
                labels.length, 2,
                2, 20,
                20, 15);
        return formPanel;
    }

    private JPanel getMainPanel() {
        return mainPanel;
    }

    public void setOkListener(OkListener okListener) {
        this.okListener = okListener;
    }

    private void showMessage(String message){
        DialogUtil.displayInfoDialog(mainPanel, message);
    }

    private static boolean isInteger(String value){
        try{
            Integer.parseInt(value);
            return true;
        } catch (NumberFormatException e){
            return false;
        }
    }

    private static boolean isFloat(String dbl) {
        try{
            Double.parseDouble(dbl);
            return true;
        } catch (NumberFormatException e){
            return false;
        }
    }

    @Override
    public void display() {
        dialog.setVisible(true);
    }

    @Override
    public void dispose() {
        dialog.dispose();
    }
}
